#include <stdio.h>
#include <stdlib.h>
#include <stdarg.h>
#include <ctype.h>
#include <memory.h>
#include <string.h>

#include <w32api/windows.h>
#include <time.h>
#include "ftd2xx.h"

#define N_ELEMS(A) (sizeof(A)/sizeof((A)[0]))

static void error_exit(const char* fmt, ...)
{
  va_list ap;
  va_start(ap, fmt);
  vfprintf(stderr, fmt, ap);
  va_end(ap);
  exit(1);
}


//*****************************************************************************
// Declare USB procedures. This part is application independent.


static const char* getFtErrorMessage(FT_STATUS ftStatus);

// Enumerate procedure informations as triples
// (Procedure name, param types, param names). 

#define USB_PROCS \
  \
PROC(FT_Open, \
     (int iDevice, \
      FT_HANDLE *pftHandle), \
     (iDevice, pftHandle)) \
PROC(FT_ListDevices, \
     (PVOID pvArg1, \
      PVOID pvArg2, \
      DWORD dwFlags), \
     (pvArg1, pvArg2, dwFlags)) \
PROC(FT_Close, \
     (FT_HANDLE ftHandle), \
     (ftHandle)) \
PROC(FT_Read, \
     (FT_HANDLE ftHandle, \
      LPVOID lpBuffer, \
      DWORD dwBytesToRead, \
      LPDWORD lpdwBytesReturned), \
     (ftHandle, lpBuffer, dwBytesToRead, lpdwBytesReturned)) \
PROC(FT_Write, \
     (FT_HANDLE ftHandle, \
      LPVOID lpBuffer, \
      DWORD dwBytesToWrite, \
      LPDWORD lpdwBytesWritten), \
     (ftHandle, lpBuffer, dwBytesToWrite, lpdwBytesWritten)) \
PROC(FT_ResetDevice, \
     (FT_HANDLE ftHandle), \
     (ftHandle)) \
PROC(FT_SetTimeouts, \
     (FT_HANDLE ftHandle, \
      DWORD dwReadTimeout, \
      DWORD dwWriteTimeout), \
     (ftHandle, dwReadTimeout, dwWriteTimeout)) \
PROC(FT_GetStatus, \
     (FT_HANDLE ftHandle, \
      LPDWORD lpdwAmountInRxQueue, \
      LPDWORD lpdwAmountInTxQueue, \
      LPDWORD lpdwEventStatus), \
     (ftHandle, lpdwAmountInRxQueue, lpdwAmountInTxQueue, lpdwEventStatus))

// enumerate procedure IDs
#define PROC(NAME, PARAMTYPES, PARAMLIST) ID_##NAME, 
enum { USB_PROCS FT_MAX };
#undef PROC


static void *ftFunctions[FT_MAX]; // pointers to USB procedures
static HMODULE usbDll = NULL;     // handle of USB DLL


// enumerate procedure names
#define PROC(NAME, PARAMTYPES, PARAMLIST) #NAME ,
static const char *usbProcNames[] = { USB_PROCS };
#undef PROC

// declare procedure types
#define PROC(NAME, PARAMTYPES, PARAMLIST) \
  typedef FTD2XX_API FT_STATUS (WINAPI *TYPE_##NAME) PARAMTYPES;
USB_PROCS
#undef PROC


//generate wrapper function of each procedure
#define PROC(NAME, PARAMTYPES, PARAMLIST) \
  FT_STATUS USB_##NAME PARAMTYPES { \
    return (*(TYPE_##NAME)(ftFunctions[ID_##NAME])) PARAMLIST; \
  }
USB_PROCS
#undef PROC

// generate each procedure's wrapper function with error check.
#define PROC(NAME, PARAMTYPES, PARAMLIST) \
  FT_STATUS EC_USB_##NAME PARAMTYPES { \
    FT_STATUS ftStatus = USB_##NAME PARAMLIST; \
    if(!FT_SUCCESS(ftStatus)){ \
      fprintf(stderr, "Error : " #NAME "failed : %s\n", \
              getFtErrorMessage(ftStatus)); \
      exit(1); \
    } \
    return ftStatus; \
  }
USB_PROCS
#undef PROC


//-----------------------------------------------------------------------------
// returns an error message coressponds to the ftStatus

static const char* getFtErrorMessage(FT_STATUS ftStatus)
{
  static const char* ftErrorMessages[] = {
    "No error",
    "Invalid Handle",
    "Device not found",
    "Device not opened",
    "I/O error",
    "Insufficient resouces",
    "Invalid parameter",
  };
  
  if(((UINT)ftStatus) >= N_ELEMS(ftErrorMessages)){
    return "Unknown Error";
  } else {
    return ftErrorMessages[ftStatus];
  } 
}

//-----------------------------------------------------------------------------
// setup the USB DLL library

static void setupDll()
{
  unsigned u;
  const char* filename = "Ftd2xx.dll";
  usbDll = LoadLibrary(filename);
  if(usbDll == NULL) {
    fprintf(stderr, "Error: Can't Load %s\n", filename);
    exit(1);
  }
	
  for(u = 0; u < N_ELEMS(usbProcNames); u++){
    void *proc = (void*)GetProcAddress(usbDll, usbProcNames[u]);
    if(proc == NULL){
      fprintf(stderr, "Error : Can't Find %s\n", usbProcNames[u]);
      exit(1);
    }
    ftFunctions[u] = proc;
  }
}

//-----------------------------------------------------------------------------
// clean up the USB DLL library

static void cleanupDll()
{
  if(usbDll != NULL){
    FreeLibrary(usbDll);
    usbDll = NULL;
  }
}

const  FT_HANDLE FT_NULL_HANDLE = (FT_HANDLE)-1;
static FT_HANDLE ftHandle = FT_NULL_HANDLE;

//-----------------------------------------------------------------------------
// close the USB device and clean up the USB DLL

static void cleanApp()
{
  if(ftHandle != FT_NULL_HANDLE){
    USB_FT_ResetDevice(ftHandle);
    USB_FT_Close(ftHandle);
    ftHandle = FT_NULL_HANDLE;
  }
  cleanupDll();
}

//-----------------------------------------------------------------------------
// initailze and open the USB device

static void initUSB()
{
  setupDll();
  atexit(cleanApp);
  
  EC_USB_FT_Open(0, &ftHandle);
  fprintf(stderr, "Open OK, handle is %p\n", ftHandle);

  // Reset and wait for the USB device to settle down
  EC_USB_FT_ResetDevice(ftHandle);
  Sleep(1000);  // wait for 1000ms
}

//*****************************************************************************
// a getc primitives of the USB


static unsigned char usb_rbuf[16384];             // receive buffer
static unsigned char* usb_rbuf_start = usb_rbuf; // start of the valid data
static unsigned char* usb_rbuf_end = usb_rbuf;   // end(+1) of the valid data

static int ungetc_char = EOF;                    // ungetc data

//-----------------------------------------------------------------------------
// receive characters from the USB and store them into the usb_rbuf

static void usb_fill_buffer(void)
{
  DWORD rq_count, tq_count, stat, n_read;

  for(;;){
    EC_USB_FT_GetStatus(ftHandle, &rq_count, &tq_count, &stat);
    if(rq_count !=0)
      break;
    Sleep(1);
  }
  if(rq_count > sizeof(usb_rbuf))
    rq_count = sizeof(usb_rbuf);
  
  EC_USB_FT_Read(ftHandle, usb_rbuf, rq_count, &n_read);
  usb_rbuf_start = usb_rbuf;
  usb_rbuf_end = usb_rbuf + n_read;
}

// output PPM file stream

static FILE* ppmout = NULL;

//-----------------------------------------------------------------------------
// receive a character from the USB
// RETURN value: the next character from the USB device
static int usb_getc(void)
{
  int c;
  if(ungetc_char != EOF){
    c = ungetc_char;
    ungetc_char = EOF;
    return c;
  }
  if(usb_rbuf_start >= usb_rbuf_end)
    usb_fill_buffer();
  
  c = *(usb_rbuf_start++);

  // write the received character to the output PPM file stream when it is active
  if(ppmout)
    fputc(c, ppmout);
  
  return c;
}

//-----------------------------------------------------------------------------
// push a character back onto the USB stream
// c : the character to be pushed back.
static void usb_ungetc(int c)
{
  ungetc_char = c;
}

//*****************************************************************************
// PPM parser
//*****************************************************************************

// PPM header information
struct ppm_info {
  int version; // must be 3 or 6
  int width;
  int height;
  int max;     // defined but not used
};

//-----------------------------------------------------------------------------
// receive an integer value in text format
// RETURN value : -1     -> error
//                others -> the integer read from the USB device
static int ppm_recv_text_number()
{
  int n = 0;
  int c = EOF;
  int prev;
  bool integer_detected = false;
  
  // skip preceding characters
  for(;;){
    
    do {  // skip spaces
      prev = c;
      c = usb_getc();
    } while(isspace(c));

    if(c == '#') {  // skip a comment line
      do {
        c = usb_getc();
        if(c == EOF)
          return -1;
      } while(c != '\n');
    } else
      break;
  }
  // parse an integer
  while(isdigit(c)){
    n = n*10 + (c - '0');
    c = usb_getc();
    integer_detected = true;
  }
  usb_ungetc(c);
  return integer_detected ? n : -1;
}

//-----------------------------------------------------------------------------
// receive a PPM header
// pi : ptr to a ppm_info struct where the header information will be stored.
// RETURN value : true  -> success
//                false -> error
static bool ppm_recv_header(ppm_info* pi)
{
  ppm_info p;

  // version (must be P3 or P6)
  if(usb_getc() != 'P') return false;

  switch(usb_getc()){
  case '3': p.version = 3; break;
  case '6': p.version = 6; break;
  default: return false;
  }
  // width
  if((p.width = ppm_recv_text_number()) < 0)
    return false;
  // height
  if((p.height = ppm_recv_text_number()) < 0)
    return false;
  // max value (read but not used)
  if((p.max = ppm_recv_text_number()) < 0)
    return false;
  // header must end with a space character
  if(!isspace(usb_getc()))
    return false;
  *pi = p;
  return true;
}

//-----------------------------------------------------------------------------
// receive a RGB pixel in text format
// buf[3] : an array where the RGB value is stored.
// RETURN value : true  -> success
//                false -> error
static bool ppm_recv_text_pixel(unsigned char buf[])
{
  int r, g, b;
  if((r = ppm_recv_text_number()) < 0) return false;
  if((g = ppm_recv_text_number()) < 0) return false;
  if((b = ppm_recv_text_number()) < 0) return false;
  buf[0] = (unsigned char)r;
  buf[1] = (unsigned char)g;
  buf[2] = (unsigned char)b;
  return true;
}

//-----------------------------------------------------------------------------
// receive a RGB pixel in binary format
// buf[3] : an array where the RGB value is stored.
// RETURN value : true  -> success
//                false -> error
static bool ppm_recv_bin_pixel(unsigned char buf[])
{
  int i;
  unsigned char buf2[3];
  
  for(i = 0; i < 3; i++){
    int c = usb_getc();
    //if(c == EOF)
    //  return false;
    buf2[i] = (unsigned char)c;
  }
  memcpy(buf, buf2, 3);
  return true;
}

//-----------------------------------------------------------------------------
// receive a RGB pixel
// buf[3] : an array where the RGB value is stored.
// version : PPM version; must be 3 or 6.
// RETURN value : true  -> success
//                false -> error
static bool ppm_recv_pixel(unsigned char buf[], int version)
{
  switch(version){
  case 3: return ppm_recv_text_pixel(buf);
  case 6: return ppm_recv_bin_pixel(buf);
  default: return false;
  }
}


//*****************************************************************************
// SLD Reader

// convert the input SLD text file into a binary format

static const unsigned MAX_N_WORDS = 4096;

union fi_union {
  int i;
  float f;
};

// the binary image of the input SLD data
static fi_union sld_words[MAX_N_WORDS];
static unsigned sld_n_words = 0;

//-----------------------------------------------------------------------------
// read a float in the SLD file and append it to the array sld_words.
// fp : input SLD file stream
// RETURN value : the float read from the file
static float read_float(FILE* fp)
{
  float f;

  if(sld_n_words >= MAX_N_WORDS)
    error_exit("read_float : too many sld words.\n");

  if(fscanf(fp, "%f", &f) != 1)
    error_exit("failed to read a float\n");

  return (sld_words[sld_n_words++].f = f);
}

//-----------------------------------------------------------------------------
// read an integer in the SLD file and append it to the array sld_words.
// fp : input SLD file stream
// RETURN value : the integer read from the file
static int read_int(FILE* fp)
{
  int i;
  if(sld_n_words >= MAX_N_WORDS)
    error_exit("read_int : too many sld words.\n");

  if(fscanf(fp, "%d", &i) != 1)
    error_exit("failed to read an int\n");
  
  return (sld_words[sld_n_words++].i = i);
}

//-----------------------------------------------------------------------------
// read a 3D float vector and append it to the array sld_words.
// fp : input SLD file stream
static void read_vec3(FILE* fp)
{
  read_float(fp);
  read_float(fp);
  read_float(fp);
}

//-----------------------------------------------------------------------------
// read the scene environments
// fp : input SLD file stream
static void read_sld_env(FILE* fp)
{
  // screen pos
  read_vec3(fp);
  // screen rotation
  read_float(fp);  read_float(fp); 
  // n_lights : Actually, it should be an int value !
  read_float(fp);
  // light rotation
  read_float(fp);  read_float(fp);
  // beam  
  read_float(fp);
}

//-----------------------------------------------------------------------------
// read all the objects
// fp : input SLD file stream
static void read_objects(FILE* fp)
{

  while (read_int(fp) != -1) {  // texture : -1 -> end
    // form
    read_int(fp);
    // refltype
    read_int(fp);
    // isrot_p
    int is_rot = read_int(fp);
    // abc
    read_vec3(fp);
    // xyz
    read_vec3(fp);
    // is_invert
    read_float(fp);
    // refl_param
    read_float(fp); read_float(fp);
    // color
    read_vec3(fp);
    // rot
    if(is_rot){
      read_vec3(fp);
    }
  }
}

//-----------------------------------------------------------------------------
// read the AND-network
// fp : input SLD file stream
static void read_and_network(FILE* fp)
{
  while(read_int(fp) != -1){
    while(read_int(fp) != -1);
  }
}

//-----------------------------------------------------------------------------
// read the OR-network
// fp : input SLD file stream
static void read_or_network(FILE* fp)
{
  while(read_int(fp) != -1){
    while(read_int(fp) != -1);
  }
}

//-----------------------------------------------------------------------------
// read the SLD file
// fp : input SLD file stream
static void read_sld(FILE* fp, bool conv_to_big_endian)
{
  read_sld_env(fp);
  read_objects(fp);
  read_and_network(fp);
  read_or_network(fp);
  if(conv_to_big_endian){
    unsigned u;
    for(u = 0; u < sld_n_words; u++){
      int i = sld_words[u].i;
      sld_words[u].i = 
        ((i & 0xff) << 24) | ((i & 0xff00) << 8) |
          ((i >> 8) & 0xff00) | ((i >> 24) & 0xff);
    }
  }
}

//*****************************************************************************
// Each step of the server

static DWORD startTime;
static DWORD endTime;

//-----------------------------------------------------------------------------
// read the SLD file and convert into a binary format.
// sld_file_name : name of the SLD data file
static void load_sld_file(const char* sld_file_name, bool conv_to_big_endian)
{
  FILE* fp = sld_file_name ? fopen(sld_file_name, "rb") : stdin;
  if(fp == NULL)
    error_exit("cannot open SLD file %s\n", sld_file_name);
  read_sld(fp, conv_to_big_endian);

  fclose(fp);
}

//-----------------------------------------------------------------------------
// send the initialization data
// init_file_name : name of the initialization data file
static void send_init_data(const char* init_file_name)
{

  if(init_file_name){
    FILE *fp = fopen(init_file_name, "rb");
    if(fp == NULL)
      error_exit("cannot open initialization data file %s\n", init_file_name);
    
    unsigned n_read;
    DWORD dummy;
    unsigned total = 0;
    char buf[16384];
    fprintf(stderr, "sending initialization data:\n");
    
    while((n_read = fread(buf, 1, sizeof(buf), fp))){
      EC_USB_FT_Write(ftHandle, buf, n_read, &dummy);

      if(dummy != n_read) // must not occur in normal environment
        fprintf(stderr, "\twarning : only %d/%d bytes wrote.\n", dummy, n_read);
      
      total += n_read;
      fprintf(stderr, "\twrote %u bytes\r", total);
    }
    fprintf(stderr, "\n\tdone.\n");
    fclose(fp);
  }
}


//-----------------------------------------------------------------------------
// wait until a 0xaa byte is received
static void wait_for_0xaa(void)
{
  fprintf(stderr, "waiting for a 0xaa byte:\n");
  while(usb_getc() != 0xaa);
  fprintf(stderr, "\treceived.\n");
}


//-----------------------------------------------------------------------------
// send the SLD data in a binary format.
static void send_sld_data(void)
{
  unsigned n_write = sld_n_words*sizeof(sld_words[0]);
  DWORD dummy;
  fprintf(stderr, "sending SLD data:\n");
  startTime = GetTickCount();
  EC_USB_FT_Write(ftHandle, sld_words, n_write, &dummy);
  if(dummy != n_write) // must not occur in normal environment
    fprintf(stderr, "\twarning : only %d/%d bytes wrote.\n", dummy, n_write);
  fprintf(stderr, "\n\tdone.\n");
}

//-----------------------------------------------------------------------------
// receive the resulting PPM image from the USB device
// out : output PPM file stream
static void recv_ppm_data(FILE* out)
{
  ppm_info pi;
  fprintf(stderr, "receiving PPM data:\n");
  
  ppmout = out;
  
  if(!ppm_recv_header(&pi))
    error_exit("error in PPM header\n");
  unsigned n_pixels = pi.width * pi.height;
  unsigned u;
  
  fprintf(stderr, "\tversion: %d       size: %dx%d\n",
          pi.version, pi.width, pi.height);
  
  for(u = 0; u < n_pixels; u++){
    unsigned char buf[3];
    if(!ppm_recv_pixel(buf, pi.version))
      error_exit("error in receiving pixels(%d more pixel(s) expected).\n",
                 n_pixels - u);
    if((u + 1) % 128 == 0)
      fprintf(stderr, "\treceived %u/%u pixels.\r", u + 1, n_pixels);
  }
  endTime = GetTickCount();
  fprintf(stderr, "\n\tdone.\n");
}


  
static const char usage[] =
" usage : %s [options] [input SLD file(default stdin)]\n"
" options :\n"
"\t-i filename   : specify binary file for initialization(default none)\n"
"\t-o filename   : specify output PPM file name(default stdout)\n"
"\t-b            : send SLD data in big-endian format.\n"
"\t-h            : print this help.\n";

//-----------------------------------------------------------------------------
// close the PPM output stream if it is active
static void clean_app(void)
{
  if(ppmout){
    fclose(ppmout);
    ppmout = NULL;
  }
}

//-----------------------------------------------------------------------------
// main
int main(int argc, char* argv[])
{
  int i;
  char* sld_file_name = NULL;
  char* init_file_name = NULL;
  char* ppm_file_name = NULL;
  bool conv_to_big_endian = false;
  atexit(clean_app);
  
  // parse args
  for(i = 1; i < argc; i++){
    
    if(!strcmp(argv[i], "-i")){
      if(++i >= argc)
        error_exit(usage, argv[0]);
      init_file_name = argv[i];

    } else if(!strcmp(argv[i], "-o")){
      if(++i >= argc)
        error_exit(usage, argv[0]);
      ppm_file_name = argv[i];

    } else if(!strcmp(argv[i], "-b")){
      conv_to_big_endian = true;

    } else if(!strcmp(argv[i], "-h")){
      fprintf(stderr, usage, argv[0]);
      return 0;

    } else if(i+1 == argc){
      sld_file_name = argv[i];

    } else {
      error_exit(usage, argv[0]);
    }
  }

  // open the output PPM file
  FILE* out = ppm_file_name ? fopen(ppm_file_name, "wb") : stdout;
  if(!out)
    error_exit("cannot open PPM output file %s\n", ppm_file_name);

  // load SLD data
  load_sld_file(sld_file_name, conv_to_big_endian);

  // send the initialization data
  initUSB();
  send_init_data(init_file_name);

  // wait for a 0xaa byte
  wait_for_0xaa();

  // send the SLD data
  send_sld_data();

  // receive the PPM image
  recv_ppm_data(out);

  DWORD elapsedTime = endTime - startTime;
  fprintf(stderr, "elapsed time : %u.%03u s\n",
          elapsedTime/1000, elapsedTime%1000);
  return 0;
}

